#import <Foundation/Foundation.h>
#import <CallKit/CallKit.h>
#import "SINCallKitMediator+Private.h"

static CXCallEndedReason SINGetCallEndedReason(SINCallEndCause cause);

@implementation SINCallKitMediator (SINCallDelegate)

- (void)callDidProgress:(id<SINCall>)call {
  NSUUID *uuid = [self.callRegistry callKitUUIDForSinchId:call.callId];
  if (uuid != nil && call.direction == SINCallDirectionOutgoing) {
    [self.provider reportOutgoingCallWithUUID:uuid startedConnectingAtDate:call.details.startedTime];
  }

  [self fanoutDelegateCallWithCallback:^(id observer) {
    if ([observer respondsToSelector:@selector(callDidProgress:)]) {
      [observer callDidProgress:call];
    }
  }];
}

- (void)callDidEstablish:(id<SINCall>)call {
  NSUUID *uuid = [self.callRegistry callKitUUIDForSinchId:call.callId];
  if (uuid != nil && call.direction == SINCallDirectionOutgoing) {
    [self.provider reportOutgoingCallWithUUID:uuid connectedAtDate:call.details.establishedTime];
  }

  [self fanoutDelegateCallWithCallback:^(id observer) {
    if ([observer respondsToSelector:@selector(callDidEstablish:)]) {
      [observer callDidEstablish:call];
    }
  }];
}

// Handle cancel/hangup event initiated by either caller or callee
- (void)callDidEnd:(id<SINCall>)call {
  call.delegate = nil;

  NSUUID *uuid = [self.callRegistry callKitUUIDForSinchId:call.callId];
  if (uuid != nil) {
    // Report end of the call to CallKit
    [self.provider reportCallWithUUID:uuid
                          endedAtDate:call.details.endedTime
                               reason:SINGetCallEndedReason(call.details.endCause)];
  }

  [self.callRegistry removeSinchCallWithId:call.callId];

  if (call.details.endCause == SINCallEndCauseError) {
    os_log_error(self.customLog,
                 "didEndCall with reason: %{public}d, error: %{public}@, Removing sinch call with ID: %{public}@",
                 (int)call.details.endCause, call.details.error == nil ? @"?" : call.details.error.localizedDescription,
                 call.callId);
  } else {
    os_log(self.customLog, "didEndCall with reason: %{public}d, Removing sinch call with ID: %{public}@",
           (int)call.details.endCause, call.callId);
  }

  [self fanoutDelegateCallWithCallback:^(id observer) {
    if ([observer respondsToSelector:@selector(callDidEnd:)]) {
      [observer callDidEnd:call];
    }
  }];
}

- (void)callDidAddVideoTrack:(id<SINCall>)call {
  [self fanoutDelegateCallWithCallback:^(id observer) {
    if ([observer respondsToSelector:@selector(callDidAddVideoTrack:)]) {
      [observer callDidAddVideoTrack:call];
    }
  }];
}

- (void)callDidPauseVideoTrack:(id<SINCall>)call {
  [self fanoutDelegateCallWithCallback:^(id observer) {
    if ([observer respondsToSelector:@selector(callDidPauseVideoTrack:)]) {
      [observer callDidPauseVideoTrack:call];
    }
  }];
}

- (void)callDidResumeVideoTrack:(id<SINCall>)call {
  [self fanoutDelegateCallWithCallback:^(id observer) {
    if ([observer respondsToSelector:@selector(callDidResumeVideoTrack:)]) {
      [observer callDidResumeVideoTrack:call];
    }
  }];
}

@end

static CXCallEndedReason SINGetCallEndedReason(SINCallEndCause cause) {
  switch (cause) {
    case SINCallEndCauseError:
      return CXCallEndedReasonFailed;
    case SINCallEndCauseDenied:
      return CXCallEndedReasonRemoteEnded;
    case SINCallEndCauseHungUp:
      // This mapping is not really correct, as SINCallEndCauseHungUp is the end cause also when the local peer ended
      // the call.
      return CXCallEndedReasonRemoteEnded;
    case SINCallEndCauseTimeout:
      return CXCallEndedReasonUnanswered;
    case SINCallEndCauseCanceled:
      return CXCallEndedReasonUnanswered;
    case SINCallEndCauseNoAnswer:
      return CXCallEndedReasonUnanswered;
    case SINCallEndCauseOtherDeviceAnswered:
      return CXCallEndedReasonAnsweredElsewhere;
    case SINCallEndCauseInactive:
      // This mapping is not really correct, as SINCallEndInactive is triggered by the local peer.
      return CXCallEndedReasonRemoteEnded;

    default:
      break;
  }
  return CXCallEndedReasonFailed;
}
