#import "SINCallKitMediator.h"
#import "Config.h"
#import "SINJWT.h"
#import "SINCallRegistry.h"
#import "AudioControllerDelegate.h"

static CXCallEndedReason SINGetCallEndedReason(SINCallEndCause cause);

@interface SINCallKitMediator () <SINCallClientDelegate, SINClientDelegate, SINCallDelegate, CXProviderDelegate>
@property (nonatomic, readwrite, strong) _Nullable id<SINClient> client;
@property (nonatomic, readwrite, strong) os_log_t customLog;

@property (nonatomic, readwrite, strong) AudioControllerDelegate *acDelegate;
@property (nonatomic, readwrite, strong) SINCallRegistry *callRegistry;
@property (nonatomic, readwrite, strong) CXProvider *provider;
@property (nonatomic, readwrite, strong) CXCallController *callController;

@property (nonatomic, readwrite, strong) _Nullable ErrorCompletion clientCreatedCallback;
@property (nonatomic, readwrite, strong) _Nullable CallStartedCompletion callStartedCallback;

@property (nonatomic, strong, readonly) NSPointerArray *observers;
@end

@implementation SINCallKitMediator

- (instancetype)init {
  self = [super init];

  if (self != nil) {
    self.customLog = os_log_create("com.sinch.sdk.app", "SINCallKitMediator");
    self.callRegistry = [[SINCallRegistry alloc] init];

    CXProviderConfiguration *config = [[CXProviderConfiguration alloc] initWithLocalizedName:@"Sinch"];
    config.maximumCallGroups = 1;
    config.maximumCallsPerCallGroup = 1;
    config.ringtoneSound = @"incoming.wav";
    config.supportedHandleTypes = [NSSet setWithArray:@[ @(CXHandleTypeGeneric) ]];

    self.acDelegate = [[AudioControllerDelegate alloc] init];
    self.callController = [[CXCallController alloc] init];
    self.provider = [[CXProvider alloc] initWithConfiguration:config];
    [self.provider setDelegate:self queue:nil];
    _observers = [NSPointerArray weakObjectsPointerArray];
  }

  return self;
}

- (void)createClientWithUserId:(NSString *)userId completion:(void (^)(NSError *))completion {
  if (self.client == nil) {
    NSError *error;
    self.client = [Sinch clientWithApplicationKey:APPLICATION_KEY
                                  environmentHost:ENVIRONMENT_HOST
                                           userId:userId
                                            error:&error];

    if (!self.client && completion) {
      completion(error);
      return;
    }

    self.client.delegate = self;
    self.client.callClient.delegate = self;
    self.client.audioController.delegate = self.acDelegate;

    [self.client enableManagedPushNotifications];
    self.clientCreatedCallback = completion;

    [self.client start];
  } else {
    if (completion && self.client.isStarted) {
      completion(nil);
    }
  }
}

- (void)startOutgoingCallTo:(NSString *)destination completion:(CallStartedCompletion)completion {
  CXHandle *handle = [[CXHandle alloc] initWithType:CXHandleTypeGeneric value:destination];
  CXStartCallAction *initiateCallAction = [[CXStartCallAction alloc] initWithCallUUID:[NSUUID UUID] handle:handle];
  initiateCallAction.video = NO;

  CXTransaction *transaction = [[CXTransaction alloc] initWithAction:initiateCallAction];
  self.callStartedCallback = completion;

  [self.callController requestTransaction:transaction
                               completion:^(NSError *_Nullable error) {
                                 if (error != nil) {
                                   os_log_error(self.customLog, "Error requesting start call transaction: %{public}@",
                                                error.localizedDescription);
                                   dispatch_async(dispatch_get_main_queue(), ^{
                                     if (completion) {
                                       completion(nil, error);
                                     }
                                     self.callStartedCallback = nil;
                                   });
                                 }
                               }];
}

- (void)endCall:(id<SINCall>)call {
  NSUUID *uuid = [self.callRegistry callKitUUIDForSinchId:call.callId];
  if (uuid == nil) {
    return;
  }

  CXEndCallAction *endCallAction = [[CXEndCallAction alloc] initWithCallUUID:uuid];
  CXTransaction *transaction = [[CXTransaction alloc] initWithAction:endCallAction];

  [self.callController requestTransaction:transaction
                               completion:^(NSError *_Nullable error) {
                                 if (error != nil) {
                                   os_log_error(self.customLog, "Error requesting end call transaction: %{public}@",
                                                error.localizedDescription);
                                 }
                                 self.callStartedCallback = nil;
                               }];
}

#pragma mark - Implementation

- (void)fanoutDelegateCallWithCallback:(void (^)(id observer))callback {
  NSArray *observerObjects = [self.observers allObjects];
  for (id observer in observerObjects) {
    if (observer) {
      callback(observer);
    }
  }
}

- (void)addObserver:(id<SINCallKitMediatorObserver>)observer {
  NSArray *observerObjects = [self.observers allObjects];
  if ([observerObjects containsObject:observer])
    return;

  [self.observers addPointer:(__bridge void *)observer];
}

- (void)removeObserver:(id<SINCallKitMediatorObserver>)observer {
  for (NSUInteger i = 0; i < self.observers.count; ++i) {
    if ([self.observers pointerAtIndex:i] == (__bridge void *)observer) {
      [self.observers removePointerAtIndex:i];
      break;
    }
  }
}

#pragma mark - SINClientDelegate

- (void)client:(id<SINClient>)client requiresRegistrationCredentials:(id<SINClientRegistration>)registrationCallback {
  // WARNING: Embedding the Application Secret and use of the class SINJWT as shown here
  // should NOT be done for an application deployed to production.
  // The class SINJWT is only here to show how to construct and sign a registration token.

  NSString *jwt = [SINJWT jwtForUserRegistrationWithApplicationKey:APPLICATION_KEY
                                                 applicationSecret:APPLICATION_SECRET
                                                            userId:self.client.userId];

  [registrationCallback registerWithJWT:jwt];
}

- (void)clientDidStart:(id<SINClient>)client {
  os_log(self.customLog, "Sinch client started successfully (version: %{public}@)", [Sinch versionString]);

  if (self.clientCreatedCallback != nil) {
    self.clientCreatedCallback(nil);
    self.clientCreatedCallback = nil;
  }
}

- (void)clientDidFail:(id<SINClient>)client error:(NSError *)error {
  os_log_error(self.customLog, "Sinch client error: %{public}@", [error localizedDescription]);

  self.client = nil;

  if (self.clientCreatedCallback != nil) {
    self.clientCreatedCallback(error);
    self.clientCreatedCallback = nil;
  }
}

#pragma mark - SINCallDelegate

- (void)callDidProgress:(id<SINCall>)call {
  NSUUID *uuid = [self.callRegistry callKitUUIDForSinchId:call.callId];
  if (uuid != nil && call.direction == SINCallDirectionOutgoing) {
    [self.provider reportOutgoingCallWithUUID:uuid startedConnectingAtDate:call.details.startedTime];
  }

  [self fanoutDelegateCallWithCallback:^(id observer) {
    if ([observer respondsToSelector:@selector(callDidProgress:)]) {
      [observer callDidProgress:call];
    }
  }];
}

- (void)callDidEstablish:(id<SINCall>)call {
  NSUUID *uuid = [self.callRegistry callKitUUIDForSinchId:call.callId];
  if (uuid != nil && call.direction == SINCallDirectionOutgoing) {
    [self.provider reportOutgoingCallWithUUID:uuid connectedAtDate:call.details.establishedTime];
  }

  [self fanoutDelegateCallWithCallback:^(id observer) {
    if ([observer respondsToSelector:@selector(callDidEstablish:)]) {
      [observer callDidEstablish:call];
    }
  }];
}

// Handle cancel/hangup event initiated by either caller or callee
- (void)callDidEnd:(id<SINCall>)call {
  call.delegate = nil;

  NSUUID *uuid = [self.callRegistry callKitUUIDForSinchId:call.callId];
  if (uuid != nil) {
    // Report end of the call to CallKit
    [self.provider reportCallWithUUID:uuid
                          endedAtDate:call.details.endedTime
                               reason:SINGetCallEndedReason(call.details.endCause)];
  }

  [self.callRegistry removeSinchCallWithId:call.callId];

  if (call.details.endCause == SINCallEndCauseError) {
    os_log_error(self.customLog,
                 "didEndCall with reason: %{public}d, error: %{public}@, Removing sinch call with ID: %{public}@",
                 (int)call.details.endCause, call.details.error == nil ? @"?" : call.details.error.localizedDescription,
                 call.callId);
  } else {
    os_log(self.customLog, "didEndCall with reason: %{public}d, Removing sinch call with ID: %{public}@",
           (int)call.details.endCause, call.callId);
  }

  [self fanoutDelegateCallWithCallback:^(id observer) {
    if ([observer respondsToSelector:@selector(callDidEnd:)]) {
      [observer callDidEnd:call];
    }
  }];
}

#pragma mark - CXProviderDelegate

- (void)providerDidReset:(nonnull CXProvider *)provider {
  os_log(self.customLog, "CXProviderDelegate.providerDidReset()");

  /*
   End any ongoing calls if the provider resets, and remove them from the app's list of calls
   because they are no longer valid.
   */
  NSArray *calls = [self.callRegistry activeSinchCalls];
  [calls enumerateObjectsUsingBlock:^(id<SINCall> _Nonnull call, NSUInteger idx, BOOL *_Nonnull stop) {
    [call hangup];
  }];

  // Remove all calls from the app's list of calls.
  [self.callRegistry reset];
}

- (void)provider:(CXProvider *)provider didActivateAudioSession:(AVAudioSession *)audioSession {
  if (nil == self.client) {
    os_log_error(self.customLog,
                 "SinchClient not assigned when audio session is activating (provider:didActivate audioSession)");
    return;
  }
  [self.client.callClient provider:provider didActivateAudioSession:audioSession];
}

- (void)provider:(CXProvider *)provider didDeactivateAudioSession:(AVAudioSession *)audioSession {
  if (nil == self.client) {
    os_log_error(self.customLog,
                 "SinchClient not assigned when audio session is activating (provider:didDeactivate audioSession)");
    return;
  }
  [self.client.callClient provider:provider didDeactivateAudioSession:audioSession];
}

- (void)provider:(CXProvider *)provider performStartCallAction:(CXStartCallAction *)action {
  CallStartedCompletion callback = self.callStartedCallback;
  self.callStartedCallback = nil;

  if (self.client == nil) {
    os_log_error(self.customLog, "SinchClient not assigned when CXStartCallAction. Failing action");
    [action fail];
    if (callback) {
      NSError *error =
          [NSError errorWithDomain:@"com.app.domain"
                              code:1
                          userInfo:@{NSLocalizedDescriptionKey : @"SinchClient not assigned when CXStartCallAction."}];
      callback(nil, error);
    }
    return;
  }

  os_log(self.customLog, "provider perform action: CXStartCallAction");

  id<SINCall> call = nil;
  NSError *error = nil;

  call = [self.client.callClient callPhoneNumber:action.handle.value];
  if (call.state == SINCallStateEnded) {
    os_log(self.customLog, "Failed to create outgoing call, error: %@", [call.details.error localizedDescription]);
    [action fail];
    error = call.details.error;
  } else {
    [self.callRegistry addSinchCall:call];
    [self.callRegistry mapCallKitUUID:action.callUUID toSinchCallId:call.callId];
    [self.client.audioController configureAudioSessionForCallKitCall];
    // NOTE: Mediator is the only delegate of the call and will do the fanout of events
    call.delegate = self;

    [action fulfill];
  }

  if (callback) {
    callback(call, error);
  }
}

- (void)provider:(CXProvider *)provider performEndCallAction:(CXEndCallAction *)action {
  if (self.client == nil) {
    os_log_error(self.customLog, "SinchClient not assigned when CXEndCallAction. Failing action");
    [action fail];
    return;
  }

  id<SINCall> call = [self.callRegistry sinchCallForCallKitUUID:action.callUUID];
  if (call == nil) {
    [action fail];
    return;
  }

  os_log(self.customLog, "provider perform action: CXEndCallAction: %{public}@", call.callId);

  [call hangup];

  [action fulfill];
}

- (void)provider:(CXProvider *)provider performSetMutedCallAction:(CXSetMutedCallAction *)action {
  if (self.client == nil) {
    os_log_error(self.customLog, "SinchClient not assigned when CXSetMutedCallAction. Failing action");
    [action fail];
    return;
  }

  os_log(self.customLog, "provider perform action: CXSetMutedCallAction, uuid: %{public}@",
         action.callUUID.description);

  if (self.acDelegate.muted) {
    [self.client.audioController unmute];
  } else {
    [self.client.audioController mute];
  }

  [action fulfill];
}

@end

static CXCallEndedReason SINGetCallEndedReason(SINCallEndCause cause) {
  switch (cause) {
    case SINCallEndCauseError:
      return CXCallEndedReasonFailed;
    case SINCallEndCauseDenied:
      return CXCallEndedReasonRemoteEnded;
    case SINCallEndCauseHungUp:
      // This mapping is not really correct, as SINCallEndCauseHungUp is the end cause also when the local peer ended
      // the call.
      return CXCallEndedReasonRemoteEnded;
    case SINCallEndCauseTimeout:
      return CXCallEndedReasonUnanswered;
    case SINCallEndCauseCanceled:
      return CXCallEndedReasonUnanswered;
    case SINCallEndCauseNoAnswer:
      return CXCallEndedReasonUnanswered;
    case SINCallEndCauseOtherDeviceAnswered:
      return CXCallEndedReasonAnsweredElsewhere;
    case SINCallEndCauseInactive:
      // This mapping is not really correct, as SINCallEndInactive is triggered by the local peer.
      return CXCallEndedReasonRemoteEnded;
    default:
      break;
  }
  return CXCallEndedReasonFailed;
}
